import os
import dill
from copy import deepcopy
from collections import defaultdict

from jsondler import JsonEntry
import numpy as np
import pandas as pd
from malemba import ArrayModelBase
from catboost import CatBoost


NUM_THREADS = 4


class ModelCatBoost(ArrayModelBase):

    def __init__(self, params=None, **kwargs):
        self.model = CatBoost(params=params)
        self.cat_features = set()
        super(ModelCatBoost, self).__init__(params=params, **kwargs)

    def fit(self, X, Y, **kwargs):
        """
        :param X: list or iterator of dicts with features {feat1: v1, feat2: v2, ...}
        :param Y: list of labels
        """
        X = self.get_cat_features(X=X)
        print("got categorial features")
        X, Y, data_shape = super(ModelCatBoost, self).fit(X=X, Y=Y, **kwargs)
        print("got features array shape")
        data = self.np_array(X, data_shape, low_memory=self.low_memory)
        self.model.fit(X=pd.DataFrame(data).values,
                       y=np.array(list(Y)),
                       cat_features=sorted([self._features[feat] for feat in self.cat_features]))

    def predict(self, X, **kwargs):
        """
        :param X: list or iterator of dicts with features {feat1: v1, feat2: v2, ...}
        :return: list of dicts with labels scores
        """
        X = self.get_cat_features(X=X)
        print("got categorial features")
        X, data_shape = super(ModelCatBoost, self).predict(X=X, **kwargs)
        print("got features array shape")
        data = self.np_array(X, data_shape)
        return list(map(lambda p: dict((self.labels[i],
                                        p[i]**(1.0/self.label_weights[self.labels[i]])) for i in range(len(p))),
                        self.model.predict(data=pd.DataFrame(data).values, prediction_type="Probability")))

    def dump(self, scheme_path, **kwargs):
        if not os.path.exists(scheme_path):
            os.makedirs(scheme_path)
        model = self.__dict__.pop("model")
        meta_f = open(os.path.join(scheme_path, "meta.m"), "wb")
        dill.dump(self.__dict__, meta_f)
        meta_f.close()
        self.model = model

        self.model.save_model(fname=os.path.join(scheme_path, "model.m"), format="cbm")

    @classmethod
    def load(cls, scheme_path, params=None, **kwargs):
        model_catboost = cls(params=params, **kwargs)
        with open(os.path.join(scheme_path, "meta.m"), "rb") as meta_f:
            model_catboost.__dict__ = dill.load(meta_f)
        if params is not None:
            if model_catboost.params is not None:
                model_catboost.params.update(params)
            else:
                model_catboost.params = params
        model_catboost.model = CatBoost(params=model_catboost.params)
        model_catboost.model.load_model(fname=os.path.join(scheme_path, "model.m"), format="cbm")
        return model_catboost

    @staticmethod
    def _convert_str_to_factors():
        return False

    @property
    def num_threads(self):
        return self.params.get("thread_count", NUM_THREADS)

    def get_cat_features(self, X):
        for x in X:
            x_cat = dict()
            for feat in x:
                if type(x[feat]) in (str, np.str, np.str_):
                    self.cat_features.add(feat)
                if type(x[feat]) in (bool, np.bool, np.bool_):
                    x_cat[feat] = int(x[feat])
                else:
                    x_cat[feat] = x[feat]
            yield x_cat


class CatBoostParams(JsonEntry):

    method_name = "CatBoost"

    max_depth_key = "max_depth"
    n_estimators_key = "n_estimators"
    eta_key = "eta"
    reg_lambda_key = "reg_lambda"
    objective_key = "objective"
    thread_count_key = "thread_count"
    task_type_key = "task_type"

    max_depth_0 = 6
    n_estimators_0 = 1000
    eta_0 = 0.1
    reg_lambda_0 = 3
    objective_0 = 'MultiClass'
    thread_count_0 = NUM_THREADS
    task_type_0 = "CPU"

    def __init__(self,
                 max_depth=max_depth_0,
                 n_estimators=n_estimators_0,
                 eta=eta_0,
                 reg_lambda=reg_lambda_0,
                 objective=None,
                 thread_count=thread_count_0,
                 task_type=None):

        if task_type is None:
            task_type = self.task_type_0
        if objective is None:
            objective = self.objective_0

        self.max_depth = max_depth
        self.n_estimators = n_estimators
        self.eta = eta
        self.reg_lambda = reg_lambda
        self.objective = objective
        self.thread_count = thread_count
        self.task_type = task_type

    @classmethod
    def attr_scheme(cls):
        attr_scheme = {
            "max_depth": (cls.max_depth_key,),
            "n_estimators": (cls.n_estimators_key,),
            "eta": (cls.eta_key,),
            "reg_lambda": (cls.reg_lambda_key,),
            "objective": (cls.objective_key,),
            "thread_count": (cls.thread_count_key,),
            "task_type": (cls.task_type_key,),
        }
        return attr_scheme

    @property
    def params(self):
        return super(CatBoostParams, self).get_json()



import os
import dill
from copy import deepcopy

from malemba import ArrayModelBase
from jsondler import JsonEntry
import numpy as np
import pandas as pd
import xgboost as xgb


NUM_THREADS = 4


class ModelXGB(ArrayModelBase):

    def __init__(self, params=None, **kwargs):
        self.model = None
        super(ModelXGB, self).__init__(params=params, **kwargs)

    def fit(self, X, Y, **kwargs):
        """
        :param X: list or iterator of dicts with features {feat1: v1, feat2: v2, ...}
        :param Y: list of labels
        """
        X, Y, data_shape = super(ModelXGB, self).fit(X=X, Y=Y, **kwargs)
        data = self.np_array(X, data_shape, low_memory=self.low_memory)
        dtrain = xgb.DMatrix(data=pd.DataFrame(data).values, label=np.array(list(Y)), missing=np.nan)
        self.model = xgb.train(params=self.params, dtrain=dtrain)

    def predict(self, X, **kwargs):
        """
        :param X: list or iterator of dicts with features {feat1: v1, feat2: v2, ...}
        :return: list of dicts with labels scores
        """
        X, data_shape = super(ModelXGB, self).predict(X=X, **kwargs)
        data = self.np_array(X, data_shape)
        dpred = xgb.DMatrix(data=pd.DataFrame(data).values, missing=np.nan)
        return list(map(lambda p: dict((self.labels[i], p[i]) for i in range(len(p))),
                        self.model.predict(dpred, output_margin=True)))

    def dump(self, scheme_path, **kwargs):
        if not os.path.exists(scheme_path):
            os.makedirs(scheme_path)
        model = self.__dict__.pop("model")
        meta_f = open(os.path.join(scheme_path, "meta.m"), "wb")
        dill.dump(self.__dict__, meta_f)
        meta_f.close()
        self.model = model

        self.model.save_model(fname=os.path.join(scheme_path, "model.m"))

    @classmethod
    def load(cls, scheme_path, params=None, **kwargs):
        model_xgb = cls(params=params, **kwargs)
        with open(os.path.join(scheme_path, "meta.m"), "rb") as meta_f:
            model_xgb.__dict__ = dill.load(meta_f)
        if params is not None:
            if model_xgb.params is not None:
                model_xgb.params.update(params)
            else:
                model_xgb.params = params
        model_xgb.model = xgb.Booster(params=model_xgb.params)
        model_xgb.model.load_model(fname=os.path.join(scheme_path, "model.m"))
        return model_xgb

    @staticmethod
    def _convert_str_to_factors():
        return True

    @property
    def num_threads(self):
        return self.params.get("nthread", NUM_THREADS)


class XGBoostParams(JsonEntry):

    method_name = "XGBoost"

    max_depth_key = "max_depth"
    min_child_weight_key = "min_child_weight"
    colsample_bytree_key = "colsample_bytree"
    gamma_key = "gamma"
    eta_key = "eta"
    reg_lambda_key = "reg_lambda"
    reg_alpha_key = "reg_alpha"
    objective_key = "objective"
    nthread_key = "nthread"
    tree_method_key = "tree_method"
    silent_key = "silent"

    max_depth_0 = 6
    min_child_weight_0 = 1.0
    colsample_bytree_0 = 1.0
    gamma_0 = 0.5
    eta_0 = 0.1
    reg_lambda_0 = 3
    reg_alpha_0 = 0.0
    objective_0 = 'multi:softmax'
    nthread_0 = NUM_THREADS
    tree_method_0 = 'auto'
    silent_0 = 1

    def __init__(self,
                 max_depth=max_depth_0,
                 min_child_weight=min_child_weight_0,
                 colsample_bytree=colsample_bytree_0,
                 gamma=gamma_0,
                 eta=eta_0,
                 reg_lambda=reg_lambda_0,
                 reg_alpha=reg_alpha_0,
                 objective=None,
                 nthread=nthread_0,
                 tree_method=None,
                 silent=silent_0):

        if tree_method is None:
            tree_method = self.tree_method_0
        if objective is None:
            objective = self.objective_0

        self.max_depth = max_depth
        self.min_child_weight = min_child_weight
        self.colsample_bytree = colsample_bytree
        self.gamma = gamma
        self.eta = eta
        self.reg_lambda = reg_lambda
        self.reg_alpha = reg_alpha
        self.objective = objective
        self.nthread = nthread
        self.tree_method = tree_method
        self.silent = silent

    @classmethod
    def attr_scheme(cls):
        attr_scheme = {
            "max_depth": (cls.max_depth_key,),
            "min_child_weight": (cls.min_child_weight_key,),
            "colsample_bytree": (cls.colsample_bytree_key,),
            "gamma": (cls.gamma_key,),
            "eta": (cls.eta_key,),
            "reg_lambda": (cls.reg_lambda_key,),
            "reg_alpha": (cls.reg_alpha_key,),
            "objective": (cls.objective_key,),
            "nthread": (cls.nthread_key,),
            "tree_method": (cls.tree_method_key,),
            "silent": (cls.silent_key,),
        }
        return attr_scheme

    @property
    def params(self):
        return super(XGBoostParams, self).get_json()



import os
import dill
from copy import deepcopy
from itertools import tee
from abc import ABCMeta, abstractmethod

from malemba import ModelBase
from malemba.ds_tools import ArrayHandler, group_array


class ModelDualStage(ModelBase):

    class Concatenator(object, metaclass=ABCMeta):

        @abstractmethod
        def concatenate(self, X, Y):
            return

    def __init__(self, model1, model2, concatenator, params1=None, params2=None, **kwargs):
        """

        :param model1: Boosting model object inherited from malemba.ArrayModelBase that is the first to be applied
        :param model2: Boosting model object inherited from malemba.ArrayModelBase that is the to be applied over model1
        :param concatenator: the object inherited from Concatenator class to be used to concatinate X and
             model1 prediction for model2 input
        :param params1: parameters dict for model1
        :param params2: parameters dict for model2
        :param kwargs:
        """

        self.model1 = model1
        self.model2 = model2
        assert isinstance(self.model1, ModelBase), \
            "Error: a value for model1 argument must be an object inherited from malemba.ModelBase"
        assert isinstance(self.model2, ModelBase), \
            "Error: a value for model2 argument must be an object inherited from malemba.ModelBase"
        self._model1_class = self.model1.__class__
        self._model2_class = self.model2.__class__
        self.concatenator = concatenator
        assert isinstance(concatenator, ModelDualStage.Concatenator), \
            "Error: a value for 'concatenator' argument must be an object inherited from ModelDualStage.Concatenator"
        self.params1 = params1
        self.params2 = params2
        super(ModelDualStage, self).__init__(params=None, **kwargs)
        self.aggr_level1 = kwargs.get("aggr_level1", 0)
        self.aggr_level2 = kwargs.get("aggr_level2", 0)
        self._data_l = 0
        self._model1_fract = 0.0

    def fit(self, X, Y, model1_fract=0.0, aggr_level1=None, aggr_level2=None, dump_scheme_path=None, **kwargs):
        """

        :param X: features data
        :param Y: labels data
        :param model1_fract: the fraction of data is to be used for model1 training
        :param aggr_level1: number of aggregation steps to turn the input data into appropriate for model1 input
        :param aggr_level2: number of aggregation steps to turn the input data into appropriate for model2 input
        :param dump_scheme_path: The path for models dump
        :param kwargs:
        :return:
        """
        super(ModelDualStage, self).fit(X=X, Y=Y, **kwargs)

        if aggr_level1 is not None:
            self.aggr_level1 = aggr_level1
        if aggr_level2 is not None:
            self.aggr_level2 = aggr_level2

        if model1_fract > 0.0:
            self._model1_fract = model1_fract
            self._data_l = kwargs.get("data_l", 0)
            if self._data_l <= 0:
                Y, Y_l = tee(Y)
                for y in Y_l:
                    self._data_l += 1

            X_1 = ModelDualStage._data_part(X, int(float(self._data_l) * self._model1_fract))
            Y_1 = ModelDualStage._data_part(Y, int(float(self._data_l) * self._model1_fract))
            if self.aggr_level1 > 0:
                data1_handler = ArrayHandler()
                Y_1 = data1_handler.aggregate(Y_1, aggr_level=self.aggr_level1)
                data1_handler.group_lims = None
                X_1 = data1_handler.aggregate(X_1, aggr_level=self.aggr_level1)
            self.model1.fit(X=X_1, Y=Y_1)
            if dump_scheme_path is not None:
                try:
                    self.model1.dump(scheme_path=os.path.join(dump_scheme_path, "model1"))
                except:
                    print("WARNING: model1 dump failed")

        if self._model1_fract >= 1.0:
            print("WARNING: model1 fraction >= 1.0 - model2 is model1")
            self.model2 = self.model1
        else:
            X = self._get_model2_data(X=X, aggr_level=self.aggr_level1)
            if self.aggr_level2 > 0:
                data2_handler= ArrayHandler()
                Y = data2_handler.aggregate(Y, aggr_level=self.aggr_level2)
                data2_handler.group_lims = None
                X = data2_handler.aggregate(X, aggr_level=self.aggr_level2)
            self.model2.fit(X=X, Y=Y)

        if dump_scheme_path is not None:
            try:
                self.model2.dump(scheme_path=os.path.join(dump_scheme_path, "model2"))
            except:
                print("WARNING: model2 dump failed")
            self.dump(scheme_path=dump_scheme_path, only_meta=True)

    @staticmethod
    def _data_part(data, part_l):
        l = 0
        for d in data:
            if l >= part_l:
                break
            l += 1
            yield d

    def predict(self, X, aggr_level1=None, aggr_level2=None, **kwargs):
        """

        :param X: features data
        :param concat_method: the function to be used to concatinate X and model1 prediction for model2 input
        :param aggr_level1: number of aggregation steps to turn the input data into one dimensional array
        :param kwargs:
        :return:
        """
        if aggr_level1 is not None:
            self.aggr_level1 = aggr_level1
        if aggr_level2 is not None:
            self.aggr_level2 = aggr_level2

        X = self._get_model2_data(X=X, aggr_level=self.aggr_level1)
        if self.aggr_level2 > 0:
            X_handler = ArrayHandler()
            X = X_handler.aggregate(X, aggr_level=self.aggr_level2)
            return group_array(self.model2.predict(X=X), group_lims=X_handler.group_lims)
        else:
            return self.model2.predict(X=X)

    def _get_model2_data(self, X, aggr_level=0):
        X, X_1 = tee(X)
        if aggr_level > 0:
            X_1_handler = ArrayHandler()
            X_1 = X_1_handler.aggregate(X_1, aggr_level=aggr_level)
            model1_pred = group_array(self.model1.predict(X=X_1), group_lims=X_1_handler.group_lims)
        else:
            model1_pred = self.model1.predict(X=X_1)

        return self.concatenator.concatenate(X, model1_pred)

    @staticmethod
    def _convert_str_to_factors():
        return False

    def str_to_factors(self, X):
        pass

    def validate(self, X_test, Y_test, labels_to_remove=None, aggr_level1=None, aggr_level2=None):
        if aggr_level1 is not None:
            self.aggr_level1 = aggr_level1
        if aggr_level2 is not None:
            self.aggr_level2 = aggr_level2

        X_test = self._get_model2_data(X=X_test, aggr_level=self.aggr_level1)
        if self.aggr_level1 > 0:
            data_handler = ArrayHandler()
            Y_test = data_handler.aggregate(Y_test, aggr_level=self.aggr_level2)
            data_handler.group_lims = None
            X_test = data_handler.aggregate(X_test, aggr_level=self.aggr_level2)
        return self.model2.validate(X_test=X_test, Y_test=Y_test, labels_to_remove=labels_to_remove)

    def validate_model1(self, X_test, Y_test, labels_to_remove=None, aggr_level=None):
        if aggr_level is not None:
            self.aggr_level1 = aggr_level

        if self.aggr_level1 > 0:
            data_handler = ArrayHandler()
            Y_test = data_handler.aggregate(Y_test, aggr_level=self.aggr_level1)
            data_handler.group_lims = None
            X_test = data_handler.aggregate(X_test, aggr_level=self.aggr_level1)
        return self.model1.validate(X_test=X_test, Y_test=Y_test, labels_to_remove=labels_to_remove)

    def validate_model2(self, X_test, Y_test, labels_to_remove=None, aggr_level1=None, aggr_level2=None):
        return self.validate(X_test=X_test,
                             Y_test=Y_test,
                             labels_to_remove=labels_to_remove,
                             aggr_level1=aggr_level1,
                             aggr_level2=aggr_level2)

    def dump(self, scheme_path, **kwargs):
        if not os.path.exists(scheme_path):
            os.makedirs(scheme_path)
        model1 = self.__dict__.pop("model1")
        model2 = self.__dict__.pop("model2")
        meta_f = open(os.path.join(scheme_path, "meta.m"), "wb")
        dill.dump(self.__dict__, meta_f)
        meta_f.close()
        self.model1 = model1
        self.model2 = model2

        if kwargs.get("only_meta", False):
            return
        self.model1.dump(scheme_path=os.path.join(scheme_path, "model1"), **kwargs)
        self.model2.dump(scheme_path=os.path.join(scheme_path, "model2"), **kwargs)

    @classmethod
    def load(cls, scheme_path, params1=None, params2=None, **kwargs):
        with open(os.path.join(scheme_path, "meta.m"), "rb") as meta_f:
            meta_dict = dill.load(meta_f)
        if params1 is not None:
            if meta_dict["params1"] is not None:
                meta_dict["params1"].update(params1)
            else:
                meta_dict["params1"] = params1
        if params2 is not None:
            if meta_dict["params2"] is not None:
                meta_dict["params2"].update(params2)
            else:
                meta_dict["params2"] = params2
        model1 = meta_dict["_model1_class"].load(scheme_path=os.path.join(scheme_path, "model1"),
                                                 params=meta_dict["params1"],
                                                 **kwargs)
        model2 = meta_dict["_model2_class"].load(scheme_path=os.path.join(scheme_path, "model2"),
                                                 params=meta_dict["params2"],
                                                 **kwargs)
        model_dual_stage = cls(model1=model1,
                               model2=model2,
                               concatenator=meta_dict["concatenator"],
                               params1=params1,
                               params2=params2)
        model_dual_stage.__dict__.update(meta_dict)
        return model_dual_stage

    @property
    def num_threads(self):
        return self.model2.num_threads

    def get_features(self, X):
        pass

    @property
    def features(self):
        return self.model1_features

    @property
    def model1_features(self):
        return self.model1.features

    @property
    def model2_features(self):
        return self.model2.features

    @property
    def feature_types(self):
        return self.model1_feature_types

    @property
    def model1_feature_types(self):
        return self.model1.feature_types

    @property
    def model2_feature_types(self):
        return self.model2.feature_types

    @property
    def labels(self):
        return self.model1.labels

    @property
    def label_freqs(self):
        if not self._label_freqs:
            for label in self.model1.label_freqs:
                self._label_freqs[label] = self.model1.label_freqs[label] * self._model1_fract
            for label in self.model2.label_freqs:
                self._label_freqs[label] += self.model2.label_freqs[label] * (1.0-self._model1_fract)
        return self._label_freqs
